import random

URANDOM = open("/dev/urandom", "rb")


def rand_bytes(size=16):
    return URANDOM.read(size)


def shuffle(in_list):
    random.shuffle(in_list, lambda: int.from_bytes(rand_bytes(32), 'big') / (2 ** 256))


def de_bruijn(k, n):
  """
  de Bruijn sequence for alphabet k
  and subsequences of length n.
  """
  try:
      # let's see if k can be cast to an integer;
      # if so, make our alphabet a list
      _ = int(k)
      alphabet = list(map(str, range(k)))

  except (ValueError, TypeError):
      alphabet = k
      k = len(k)

  a = [0] * k * n
  sequence = []

  def db(t, p):
      if t > n:
          if n % p == 0:
              sequence.extend(a[1:p + 1])
      else:
          a[t] = a[t - p]
          db(t + 1, p)
          for j in range(a[t - p] + 1, k):
              a[t] = j
              db(t + 1, t)
  db(1, 1)
  return "".join(alphabet[i] for i in sequence)



from samson.block_ciphers.modes.cbc import CBC

class CBCMAC(object):
    def __init__(self, key, cipher, iv=b'\x00' * 16):
        self.key = key
        self.iv = iv
        self.cipher = cipher

    
    def __repr__(self):
        return f"<CBCMAC: key={self.key}, iv={self.iv}, cipher={self.cipher}>"


    def __str__(self):
        return self.__repr__()

    
    def generate(self, plaintext, pad=True):
        cryptor = self.cipher(self.key)
        cbc = CBC(cryptor.encrypt, cryptor.decrypt, self.iv, cryptor.block_size // 8)
        return cbc.encrypt(plaintext, pad)[-(cryptor.block_size // 8):]


from samson.utilities.general import rand_bytes

class MarkovState(object):
    def __init__(self, count, probability, transitions):
        self.count = count
        self.probability = probability
        self.transitions = transitions


    def __repr__(self):
        return "<MarkovState: transitions={}, count={}, probability={}>".format(self.transitions.keys(), self.count, self.probability)

    def __str__(self):
        return self.__repr__()



    # Recursively turn counts into probabilities
    def calculate_chain_probs(self):
        total_count = sum([subchain.count for token, subchain in self.transitions.items()])

        for _token, subchain in self.transitions.items():
            subchain.probability = subchain.count / total_count
            subchain.calculate_chain_probs()



    def random_walk(self, distance):
        result = []
        if distance > 0:
            rand_num = int.from_bytes(rand_bytes(1), 'big')
            winning_probability = rand_num / 256
            
            prob_accumulator = 0.0

            for token, subchain in self.transitions.items():
                prob_accumulator += subchain.probability

                if prob_accumulator >= winning_probability:
                    result = [token] + subchain.random_walk(distance - 1)
                    break

        return result



from samson.prngs.bitsliced_glfsr import BitslicedGLFSR
from sympy import Poly
from sympy.abc import x

# https://github.com/ttsou/airprobe/blob/master/A5.1/python/A51_Tables/a51.py
class A51(object):
    def __init__(self, key, frame_num):
        self.key = key
        self.frame_num = frame_num

        self.lfsr_regs = [
            BitslicedGLFSR(19, 10, [13, 16, 17, 18]),
            BitslicedGLFSR(22, 11, [20, 21]),
            BitslicedGLFSR(23, 12, [7, 20, 21, 22])
        ]

        for lfsr in self.lfsr_regs:
            lfsr.mix_state(self.key, 64)
            lfsr.mix_state(self.frame_num, 22)

        for _ in range(100):
            self.clock()

       
       
    def __repr__(self):
        return f"<A51: key={self.key}, frame_num={self.frame_num}, lfsr_regs={self.lfsr_regs}>"


    def __str__(self):
        return self.__repr__()



    def clock(self):
        majority = sum([lfsr.clock_value() for lfsr in self.lfsr_regs]) // 2

        for lfsr in self.lfsr_regs:
            if lfsr.clock_value() == majority:
                lfsr.clock()


    
    def yield_state(self, size):
        bitstring = ''
        for _ in range(size):
            self.clock()
            bitstring += str(self.lfsr_regs[0].value() ^ self.lfsr_regs[1].value() ^ self.lfsr_regs[2].value())
        
        return int(bitstring, 2)


from samson.utilities.bytes import Bytes

class DGHV(object):
    def __init__(self, p=None):
        self.p = p or Bytes.random(8).to_int()


    def __repr__(self):
        return f"<DGHV: p={self.p}>"


    def __str__(self):
        return self.__repr__()



    def encrypt(self, m):
        q = Bytes.random(8).to_int()
        r = Bytes.random(8).to_int() % (self.p // 4)
        return DGHVBit(self.p * q + 2 * r + m)

    
    def decrypt(self, c):
        return (c % self.p) % 2



class DGHVBit(int):
    def __new__(cls, val, *args, **kwargs):
        return super(DGHVBit, cls).__new__(cls, val)

    def NOT(self):
        return DGHVBit(1 + self)


    def AND(self, b):
        return DGHVBit(self * b)


    def NAND(self, b):
        return DGHVBit(self.AND(b).NOT())


    def OR(self, b):
        return self.NAND(self).NAND(b.NAND(b))


    def XOR(self, b):
        return DGHVBit(self.NOT() * b + b.NOT() * self)


    def __invert__(self):
        return self.NOT()

    
    def __or__(self, b):
        return self.OR(b)


    def __and__(self, b):
        return self.AND(b)

    
    def __xor__(self, b):
        return self.XOR(b)


from samson.utilities.analysis import generate_rc4_bias_map, RC4_BIAS_MAP
import multiprocessing
import itertools
import struct

import logging
log = logging.getLogger(__name__)

# The RC4PrependAttack uses an EncryptionOracle that prepends the payload and then encrypts the whole message with a randomly generated key.
# The attack uses specific byte biases found in the RC4 keystream and feeds the plaintext into these locations. Using the law of large numbers,
# the keystream bias will surface, and we can XOR it with the ciphertext to return the plaintext. This specific implementation uses several optmizations.
# The first is concurrency/parallelism and is self-explanatory. The second is the use of concurrent biases. We first determine which biases we can still use
# in the attack. This is necessary since we can only feed the plaintext forward, so plaintext index five is not eligible to use RC4 bias index 2. Next, we determine
# which biases are "active". If the plaintext is 20 characters long, and the current padding position is 0, then we can utilize both RC4 bias indices 2 and 16 (i.e. 1 and 15 using a zero-based-index).
# Lastly, we introduce "branches". If a byte happens to decrypted more than one and has different outcomes, we keep both. At the end, we return the Cartesian product of all
# byte decryptions. This will normally just be one.

class RC4PrependAttack(object):
    # Takes an EncryptionOracle
    def __init__(self, oracle):
        self.oracle = oracle
        self.strongest_biases = [1, 15, 31]
        

    def _encrypt_chunk(self, payload, chunk_size):
        return [self.oracle.encrypt(payload) for _ in range(chunk_size)]


    def execute(self, ciphertext_length, sample_size=2**23):
        cracked_indices = [set()for i in range(ciphertext_length)]
        cpu_count = multiprocessing.cpu_count()
        pool = multiprocessing.Pool(processes=cpu_count)

        log.debug("Running with {} cores".format(cpu_count))

        for i in range(ciphertext_length):
            log.debug("Starting iteration {}/{}".format(i, ciphertext_length))

            if len(cracked_indices[i]) > 0:
                continue

            applicable_biases = [bias for bias in self.strongest_biases if i <= bias or (ciphertext_length + i >= bias and i < bias)]
            padding_len = max(applicable_biases[0] - i, 0)
            active_biases = [bias for bias in applicable_biases if padding_len + ciphertext_length > bias]

            payload = b'\x00' * padding_len
            chunk_size = sample_size // cpu_count

            log.debug("Sampling {} ciphertexts".format(sample_size))
            random_ciphertexts = [pool.apply_async(self._encrypt_chunk, (payload, chunk_size)) for i in range(cpu_count)]
            flattened_list = [result for result_list in random_ciphertexts for result in result_list.get()]

            log.debug("Generating bias map")
            bias_map = generate_rc4_bias_map(flattened_list)

            for bias_idx in active_biases:
                cracked_indices[bias_idx - padding_len].add(RC4_BIAS_MAP[bias_idx] ^ bias_map[bias_idx][0][0])


        all_branches = itertools.product(*[list(results) for results in cracked_indices])
        return [struct.pack('B' * len(branch), *branch) for branch in all_branches]


#!/usr/bin/python3
w, n, m, r = (32, 624, 397, 31)
a = 0x9908b0df
f = 1812433253
u, d = (11, 0xFFFFFFFF)
s, b = (7, 0x9D2C5680)
t, c = (15, 0xEFC60000)
l = 18


def asint32(integer):
    return integer & d

def temper(y):
    y ^= (y >> u)
    y ^= (y << s) & b
    y ^= (y << t) & c
    y ^= (y >> l)
    return y


# Implementation of MT19937
class MT19937:

    def __init__(self, seed=0):
        self.state = [0] * n
        self.seed = seed

        # Seed the algo
        self.index = n
        self.state[0] = seed

        for i in range(1, n):
            self.state[i] = asint32(f * (self.state[i - 1] ^ self.state[i - 1] >> (w - 2)) + i)


    def __repr__(self):
        return "<MT19937: seed={}, index={}, state={}>".format(self.seed, self.index, self.state)


    def __str__(self):
        return self.__repr__()
        

    def twist(self):
        for i in range(n):
            y = asint32((self.state[i] & 0x80000000) + (self.state[(i + 1) % n] & 0x7fffffff))
            self.state[i] = self.state[(i + m) % n] ^ y >> 1

            if y & 1 == 1:
                self.state[i] ^= a

        self.index = 0


    def randint(self):
        if self.index >= n:
            self.twist()

        y = self.state[self.index]
        y  = temper(y)

        self.index += 1

        return asint32(y)


if __name__ == '__main__':
    random = MT19937(0)
    assert random.randint() == 2357136044
    for i in range(1000): random.randint()
    assert random.randint() == 1193028842



import hashlib
from samson.utilities.math import modexp
from samson.utilities.general import rand_bytes
from samson.utilities.encoding import int_to_bytes
import codecs

NIST_prime = int.from_bytes(codecs.decode("ffffffffffffffffc90fdaa22168c234c4c6628b80dc1cd129024e088a67cc74020bbea63b139b22514a08798e3404ddef9519b3cd3a431b302b0a6df25f14374fe1356d6d51c245e485b576625e7ec6f44c42e9a637ed6b0bff5cb6f406b7edee386bfb5a899fa5ae9f24117c4b1fe649286651ece45b3dc2007cb8a163bf0598da48361c55d39a69163fa8fd24cf5f83655d23dca3ad961c62f356208552bb9ed529077096966d670c354e4abc9804f1746c08ca237327ffffffffffffffff", 'hex_codec'), 'little')


class SRPServer(object):
    def __init__(self, g=2, k=3, N=NIST_prime):
        self.g = g
        self.k = k
        self.N = N
        self.accounts = {}
        self.requests = {}
        self.salt = rand_bytes(4)

        self.b = int.from_bytes(rand_bytes(4), 'little') % N
        


    def create_account(self, identity, password):
        x = int.from_bytes(hashlib.sha256(self.salt + password).digest(), 'little')
        v = modexp(self.g, x, self.N)
        self.accounts[identity] = v

    
    def respond_with_challenge(self, identity, A):
        v = self.accounts[identity]
        #salt = gen_rand_key(4)
        B = self.k * v + modexp(self.g, self.b, self.N)
        self.requests[identity] = {'A': A, 'B': B} #, 'salt': self.salt
        
        return self.salt, B


    def check_challenge(self, identity, client_hash):
        request = self.requests[identity]
        v = self.accounts[identity]
        A = request['A']

        hex_A = int_to_bytes(A, 'little')
        print(hex_A)
        hex_B = int_to_bytes(request['B'], 'little')

        uH = int.from_bytes(hashlib.sha256(hex_A + hex_B).digest(), 'little')

        p1 = (A * modexp(v, uH, self.N))
        sS = modexp(p1, self.b, self.N)

        sK = hashlib.sha256(int_to_bytes(sS, 'little')).digest()
        return hashlib.sha256(sK + self.salt).digest() == client_hash


# https://stackoverflow.com/questions/14519227/rail-fence-cipher-looking-for-a-better-solution
class RailFence(object):
    def __init__(self, num_rails):
        self.num_rails = num_rails



    def _fence(self, in_list):
        fence = [[None] * len(in_list) for n in range(self.num_rails)]
        rails = list(range(self.num_rails - 1)) + list(range(self.num_rails - 1, 0, -1))

        for n, x in enumerate(in_list):
            fence[rails[n % len(rails)]][n] = x

        return [c for rail in fence for c in rail if c is not None]



    def encrypt(self, plaintext):
        return ''.join(self._fence(plaintext))


    def decrypt(self, ciphertext):
        enc_pos = self._fence(range(len(ciphertext)))
        return ''.join([ciphertext[enc_pos.index(i)] for i in range(len(ciphertext))])
        


from samson.analyzers.english_analyzer import EnglishAnalyzer
from samson.utilities.manipulation import xor_buffs
import codecs

class XORDictionaryAttack(object):
    def __init__(self, analyzer, wordlist):
        self.analyzer = analyzer
        self.wordlist = wordlist
        
    
    def execute(self, ciphertexts, word_ranges=[2,3,4]):
        two_time = xor_buffs(*ciphertexts)

        cipher_len = len(two_time)
        trimmed_list = [word for word in self.wordlist if len(word) <= cipher_len]
        prepend_list = ['']

        last_num_processed = 0
        results = []

        for j in word_ranges:
            for i in range(j - last_num_processed):
                word_scores = []
                for prepend in prepend_list:
                    for word in trimmed_list:
                        mod_word = (prepend + ' ' + word).strip()
                        xor_result = xor_buffs((bytes(mod_word, 'utf-8') + b'\x00' * cipher_len)[:cipher_len], two_time)[:len(two_time)]
                        analysis = self.analyzer.analyze(xor_result)
                        word_scores.append((mod_word, analysis / (len(word) ** 2)))

                prepend_list = [word for word, _ in sorted(word_scores, key=lambda score: score[1], reverse=True)[:10 ** (i + 1 + last_num_processed)]]
            last_num_processed = j

            results.append(sorted(prepend_list, key=lambda word: self.analyzer.analyze(xor_buffs((bytes(word, 'utf-8') + b'\x00' * cipher_len)[:cipher_len], two_time)), reverse=True)[:10])
            
        return results


from samson.utilities.manipulation import get_blocks
from samson.utilities.padding import pkcs7_unpad
import struct

import logging
log = logging.getLogger(__name__)

class ECBPrependAttack(object):
    # Expects a StatelessBlockEncryptionOracle
    def __init__(self, oracle):
        self.oracle = oracle


    def execute(self):
        baseline = len(self.oracle.encrypt(b''))
        block_size = self.oracle.find_block_size()

        plaintexts = []
        for curr_block in range(baseline // block_size):
            log.debug("Starting iteration {}".format(curr_block))
            
            plaintext = b''
            for curr_byte in range(block_size):
                if curr_block == 0:
                    payload = ('A' * (block_size - (curr_byte + 1))).encode()
                else:
                    payload = plaintexts[-1][curr_byte + 1:]

                one_byte_short = get_blocks(self.oracle.encrypt(payload), block_size=block_size)[curr_block]

                for i in range(256):
                    curr_byte = struct.pack('B', i)
                    ciphertext = self.oracle.encrypt(payload + plaintext + curr_byte)

                    # We're always editing the first block to look like block 'curr_block'
                    if get_blocks(ciphertext, block_size=block_size)[0] == one_byte_short:
                        plaintext += curr_byte
                        break

            plaintexts.append(plaintext)
        return pkcs7_unpad(b''.join(plaintexts), block_size=block_size)



from math import log10
import os

# http://practicalcryptography.com/cryptanalysis/text-characterisation/word-statistics-fitness-measure/
class ViterbiDecoder(object):
    def __init__(self):
        ''' load a file containing ngrams and counts, calculate log probabilities '''
        self.Pw = {}
        with open(os.path.join(os.path.dirname(__file__), 'count_1w.txt'), 'r') as f:
            for line in f.readlines():
                key,count = line.split('\t') 
                self.Pw[key.upper()] = self.Pw.get(key.upper(), 0) + int(count)

        self.N = 1024908267229 ## Number of tokens

        # Calculate first order log probabilities
        for key in self.Pw.keys():
            self.Pw[key] = log10(float(self.Pw[key])/self.N)

        # Get second order word model 
        self.Pw2 = {}
        with open(os.path.join(os.path.dirname(__file__), 'count_2w.txt'), 'r') as f:
            for line in f.readlines():
                key,count = line.split('\t') 
                self.Pw2[key.upper()] = self.Pw2.get(key.upper(), 0) + int(count)

        # Calculate second order log probabilities
        for key in self.Pw2.keys():
            word1,_word2 = key.split()
            if word1 not in self.Pw: 
                self.Pw2[key] = log10(float(self.Pw2[key])/self.N)
            else: 
                self.Pw2[key] = log10(float(self.Pw2[key])/self.N) - self.Pw[word1]

        # Precalculate the probabilities we assign to words not in our dict, L is length of word
        self.unseen = [log10(10./(self.N * 10**L)) for L in range(50)]        
        
        
    # Conditional word probability    
    def cPw(self,word,prev='<UNK>'):
        if word not in self.Pw: 
            return self.unseen[len(word)]
        elif prev+' '+word not in self.Pw2: 
            return self.Pw[word]
        else: 
            return self.Pw2[prev+' '+word]
    

    def score(self,text,maxwordlen=20):
        prob = [[-99e99]*maxwordlen for _ in range(len(text))]
        strs = [['']*maxwordlen for _ in range(len(text))]

        for j in range(maxwordlen):
            prob[0][j] = self.cPw(text[:j+1])
            strs[0][j] = [text[:j+1]]

        for i in range(1,len(text)):
            for j in range(maxwordlen):
                if i+j+1 > len(text): break
                candidates = [(prob[i-k-1][k] + self.cPw(text[i:i+j+1],strs[i-k-1][k][-1]),
                               strs[i-k-1][k] + [text[i:i+j+1]] ) for k in range(min(i,maxwordlen))]

                prob[i][j], strs[i][j] = max(candidates)
                
        ends = [(prob[-i-1][i],strs[-i-1][i]) for i in range(min(len(text),maxwordlen))]
        return max(ends)


from samson.utilities.math import mod_inv

class AffineCipher(object):
    def __init__(self, a, b, alphabet='ABCDEFGHIJKLMNOPQRSTUVWXYZ ,.'):
        self.a = a
        self.b = b
        self.alphabet = alphabet

        self.char_map = {}
        for i in range(len(alphabet)):
            self.char_map[alphabet[i]] = alphabet[(a*i+b)%len(alphabet)]


        inv_a = mod_inv(a, len(alphabet))
        self.inv_char_map = {}
        for i in range(len(alphabet)):
            self.inv_char_map[alphabet[i]] = alphabet[(inv_a*(i-b))%len(alphabet)]


    def __repr__(self):
        return "<Affine: a={}, b={}, alphabet={}>".format(self.a, self.b, self.alphabet)

    def __str__(self):
        return self.__repr__()


    def encrypt(self, plaintext):
        primitive_len = len(self.alphabet[0])

        ciphertext = ''
        for i in range(0, len(plaintext), primitive_len):
            curr_symbol = plaintext[i:i + primitive_len]

            if curr_symbol in self.char_map:
                ciphertext += self.char_map[curr_symbol]
            else:
                ciphertext += curr_symbol

        return ciphertext


    def decrypt(self, ciphertext):
        primitive_len = len(self.alphabet[0])

        plaintext = ''
        for i in range(0, len(ciphertext), primitive_len):
            curr_symbol = ciphertext[i:i + primitive_len]

            if curr_symbol in self.inv_char_map:
                plaintext += self.inv_char_map[curr_symbol]
            else:
                plaintext += curr_symbol

        return plaintext


from sympy import Poly
from samson.utilities.math import berlekamp_massey

class GLFSR(object):
    def __init__(self, seed, polynomial):
        self.state = seed

        if type(polynomial) is Poly:
            polynomial = int(''.join(str(coeff) for coeff in polynomial.all_coeffs()), 2)

        self.polynomial = polynomial
        self.mask = 1
        self.wrap_around_mask = 2 ** polynomial.bit_length() - 1
        self.state &= self.wrap_around_mask


        poly_mask = polynomial
        while poly_mask:
            if poly_mask & self.mask:
                poly_mask ^= self.mask
            
            if not poly_mask:
                break
            
            self.mask <<= 1


    def __repr__(self):
        return f"<GLFSR: state={self.state}, polynomial={self.polynomial}, mask={self.mask}>"
    

    def __str__(self):
        return self.__repr__()



    def clock(self):
        self.state <<= 1
        self.state &= self.wrap_around_mask

        if self.state & self.mask:
            self.state ^= self.polynomial
            return 1

        else:
            return 0
        

    
    def reverse_clock(self, output):
        for item in output:
            if item:
                self.state ^= self.polynomial
            
            self.state >>=1
            self.state &= self.wrap_around_mask



    # TODO: Make work with arbitrary polynomials!
    @staticmethod
    def crack(output):
        # Find minimum polynomial that represents the output
        poly = berlekamp_massey(output)

        # Create new LFSR and clock in reverse
        lfsr = GLFSR(0, poly)
        lfsr.reverse_clock(output[::-1])

        # Clock forward to synchronize with output
        [(lfsr.clock(), lfsr.state) for i in range(len(output))]

        return lfsr


from samson.utilities.bitstring import Bitstring

class BitslicedGLFSR(object):
    def __init__(self, length, clock_bit, taps, seed=0):
        self.state = bin(seed)[2:].zfill(length)
        self.length = length
        self.clock_bit = clock_bit
        self.taps = taps


    def __repr__(self):
        return f"<BitslicedGLFSR: state={self.state}, length={self.length}, clockbit={self.clock_bit}, taps={self.taps}>"


    def __str__(self):
        return self.__repr__()


    def mix_state(self, in_val, size):
        bit_val = Bitstring.wrap(in_val).zfill(size)

        for bit in str(bit_val)[::-1]:
            self.clock(int(bit))


    def clock(self, bit=0):
        for tap in self.taps:
            position = self.length - tap - 1
            bit ^= int(self.state[position])
        
        self.state += '0'
        self.state = self.state[1:]
        self.state = (self.state[:self.length - 1] + str(bit))


    def value(self):
        return int(self.state[0])

    
    def clock_value(self):
        return int(self.state[self.clock_bit])


from samson.utilities.manipulation import xor_buffs

# CTR will XOR the keystream with the ciphertext to decrypt the data. By injecting known plaintext,
# we can recover that block's keystream and XOR in our own plaintext. While this may seem extraneous given
# we can already inject a payload, this particular attack lends itself to bypassing validation functions.
# Imagine if an application encrypts a URL-encoded payload with its application key and gives it to us as way of keeping state.
# The developer assumes we can't modify the ciphertext without corrupting the string, so they trust the ciphertext
# when it's received.

# CBC will XOR the bitshift of the edited cipher block
# with the next blocks. To exploit this structure, we must
# craft a payload in reverse such that it creates our desired string.
# To do this, we need a known plaintext and a desired plaintext.
# We fill the targeted block with the known plaintext.
# We XOR our desired text, 'hiya;admin=true;', with the plaintext to find the "difference".
# Finally, we XOR the difference with the original cipher block.

# comment1=cooking
# %20MCs;userdata=
# aaaaaaaaaaaaaaaa
# ;comment2=%20lik
# e%20a%20pound%20
# of%20baconPPPPPP

class XORBitflippingAttack(object):
    # Expects an EncryptionOracle
    def __init__(self, oracle, block_size=16):
        self.oracle = oracle
        self.block_size = block_size


    def execute(self, desired_injection, index=16):
        payload = 'a' * self.block_size
        ciphertext = self.oracle.encrypt(payload)

        end_of_block = index + self.block_size
        edited_cipher = bytearray(ciphertext)
        edited_cipher[index:end_of_block] = xor_buffs(xor_buffs(edited_cipher[index:end_of_block], desired_injection), payload.encode())
        return edited_cipher


import math

import logging
log = logging.getLogger(__name__)


class NaiveMDCollider(object):
    def __init__(self, construction_func, output_size):
        self.construction_func = construction_func
        self.output_size = output_size


    
    @staticmethod
    def initialize_with_known_prefixes(prefixes, iv, construction_func, output_size):
        k = math.ceil(math.log(len(prefixes), 2))

        # Find the maximum length of padded prefixes
        padding_size = math.ceil(max([len(prefix) for prefix in prefixes]) / output_size) * output_size

        prefixes = [(prefix + (b'\x00' * padding_size))[:padding_size] for prefix in prefixes]
        hashed_prefixes = [list(construction_func(iv, prefix))[-1] for prefix in prefixes]

        return NaiveMDCollider(construction_func, output_size), k, prefixes, hashed_prefixes




    def find_collision(self, p1, p2):
        input_for_p1 = b'\x00' * self.output_size
        state_to_collide = list(self.construction_func(p1, input_for_p1))[0]

        # Try a lot of values...
        for j in range(2**(self.output_size * 32)):
            attempt = int.to_bytes(j, self.output_size * 4, 'little')

            # We'll need to keep a state counter since we may not get a collision on the first chunk
            state_ctr = 0

            for state in self.construction_func(p2, attempt):
                state_ctr += 1

                if state == state_to_collide:
                    log.debug('Found collision for ({}, {})'.format(p1, p2))
                    return input_for_p1, attempt[:state_ctr * self.output_size], state_to_collide



from samson.utilities.manipulation import left_rotate, get_blocks
from samson.utilities.encoding import int_to_bytes
from copy import deepcopy
from math import ceil

# from samson.utilities
# https://en.wikipedia.org/wiki/Salsa20


def quarter_round(a, b, c, d):
    a = (a + b) & 0xFFFFFFFF; d ^= a; d = left_rotate(d, 16)
    c = (c + d) & 0xFFFFFFFF; b ^= c; b = left_rotate(b, 12)
    a = (a + b) & 0xFFFFFFFF; d ^= a; d = left_rotate(d, 8)
    c = (c + d) & 0xFFFFFFFF; b ^= c; b = left_rotate(b, 7)
    return a, b, c, d


class ChaCha(object):
    def __init__(self, key, nonce, rounds=20, constant=b"expand 32-byte k"):
        self.key = key
        self.nonce = nonce
        self.rounds = rounds
        self.constant = constant



    def yield_state(self, start_chunk=0, num_chunks=1):
        for iteration in range(start_chunk, num_chunks):
            ctr_bytes = int.to_bytes(iteration, 4, 'little')

            x = [
                *[int.from_bytes(block, 'little') for block in get_blocks(self.constant, 4)],
                *[int.from_bytes(block, 'little') for block in get_blocks(self.key, 4)],
                int.from_bytes(ctr_bytes, 'little'),
                *[int.from_bytes(block, 'little') for block in get_blocks(self.nonce, 4)]
            ]


            tmp = deepcopy(x)

            for _ in range(self.rounds // 2):
                # Odd round
                x[0], x[4], x[8], x[12] = quarter_round(x[0], x[4], x[8], x[12])
                x[1], x[5], x[9], x[13] = quarter_round(x[1], x[5], x[9], x[13])
                x[2], x[6], x[10], x[14] = quarter_round(x[2], x[6], x[10], x[14])
                x[3], x[7], x[11], x[15] = quarter_round(x[3], x[7], x[11], x[15])

                # Even round
                x[0], x[5], x[10], x[15] = quarter_round(x[0], x[5], x[10], x[15])
                x[1], x[6], x[11], x[12] = quarter_round(x[1], x[6], x[11], x[12])
                x[2], x[7], x[8], x[13] = quarter_round(x[2], x[7], x[8], x[13])
                x[3], x[4], x[9], x[14] = quarter_round(x[3], x[4], x[9], x[14])

            for i in range(16):
                x[i] += tmp[i]

            yield b''.join([int_to_bytes(state_int & 0xFFFFFFFF, 'little') for state_int in x])



assert(quarter_round(0x11111111, 0x01020304, 0x9b8d6f43, 0x01234567) == (0xea2a92f4, 0xcb1cf8ce, 0x4581472e, 0x5881c4bb))
import codecs
chacha = ChaCha(key=b'\x00' * 32, nonce=b'\x00' * 12)
keystream_chunks = list(chacha.yield_state(0, 2))

assert codecs.encode(keystream_chunks[0], 'hex_codec') == b'76b8e0ada0f13d90405d6ae55386bd28bdd219b8a08ded1aa836efcc8b770dc7da41597c5157488d7724e03fb8d84a376a43b8f41518a11cc387b669b2ee6586'
assert codecs.encode(keystream_chunks[1], 'hex_codec') == b'9f07e7be5551387a98ba977c732d080dcb0f29a048e3656912c6533e32ee7aed29b721769ce64e43d57133b074d839d531ed1f28510afb45ace10a1f4b794d6f'


from samson.utilities.manipulation import get_blocks
from samson.utilities.padding import md_pad

class MerkleDamgardConstruction(object):
    def __init__(self, initial_state, compression_func, pad_func=lambda msg: md_pad(msg), block_size=64):
        self.pad_func = pad_func
        self.initial_state = initial_state
        self.compression_func = compression_func
        self.block_size = block_size


    def yield_state(self, message):
        state = self.initial_state

        for block in get_blocks(self.pad_func(message), self.block_size):
            state = self.compression_func(block, state)
            yield state


    def hash(self, message):
        final_state = list(self.yield_state(message))[-1]
        return final_state


from samson.utilities.padding import md_pad
from samson.hashes.sha1 import SHA1
import struct

def _build_sha1_internal_state(hash_bytes):
    return [struct.unpack('>I', hash_bytes[i * 4: (i + 1) * 4])[0] for i in range(len(hash_bytes) // 4)]


def _sha1_length_extension(original, message, append_bytes, secret_len):
    chunks = _build_sha1_internal_state(original)
    glue = md_pad(message, len(message) + secret_len, 'big')[len(message):]

    fake_len = secret_len + len(message) + len(glue) + len(append_bytes)

    new_hash = SHA1(chunks)
    new_hash.pad_func = lambda msg: md_pad(msg, fake_len, 'big')

    return message + glue + append_bytes, new_hash.hash(append_bytes)


class HMACForgeryAttack(object):
    def __init__(self):
        pass


    def execute(self, original_signature, message, desired_injection, secret_len):
        return _sha1_length_extension(original_signature, message, desired_injection, secret_len)



from samson.utilities.bytes import Bytes
from samson.hashes.sha1 import SHA1

def MGF1(seed, length):
    seed_int = seed.int()
    mask = b''
    sha1 = SHA1()

    for i in range(seed_int, seed_int + (length + 19) // 20):
        mask += sha1.hash(Bytes(i))
    
    return mask[:length]


# https://www.ietf.org/rfc/rfc3447.txt
class OAEP(object):
    def __init__(self, modulus_len, mgf=MGF1, hash_obj=SHA1(), label=b''):
        self.mgf = mgf
        self.hash_obj = hash_obj
        self.label = label
        self.modulus_len = modulus_len


    def __repr__(self):
        return f"<OAEP: mgf={self.mgf}, hash_obj={self.hash_obj}, label={self.label}>"


    def __str__(self):
        return self.__repr__()

        

    def pad(self, plaintext):
        plaintext = Bytes.wrap(plaintext)
        k = self.modulus_len // 8

        # Step 1: Length checking
        h_len = self.hash_obj.digest_size
        m_len = len(plaintext)
        ps_len = k - m_len - (2 * h_len) - 2

        if ps_len < 0:
            raise ValueError("Plaintext is too long")
        

        # Step 2: EME-OAEP encoding
        l_hash = self.hash_obj.hash(self.label)
        ps = Bytes(b'').zfill(ps_len)
        db = l_hash + ps + b'\x01' + plaintext

        seed = Bytes.random(h_len)

        db_mask = self.mgf(seed, k - h_len - 1)
        masked_db = db ^ db_mask

        seed_mask = self.mgf(masked_db, h_len)
        masked_seed = seed ^ seed_mask

        return b'\x00' + masked_seed + masked_db



    def unpad(self, plaintext):
        plaintext = Bytes.wrap(plaintext)
        k = self.modulus_len // 8
        h_len = self.hash_obj.digest_size
        
        masked_seed, masked_db = plaintext[1:(h_len + 1)], plaintext[(h_len + 1):]

        seed_mask = self.mgf(masked_db, h_len)
        seed = masked_seed ^ seed_mask

        db_mask = self.mgf(seed, k - h_len - 1)
        db = masked_db ^ db_mask

        l_hash, m = db[:h_len], db[h_len + db[h_len:].index(b'\x01') + 1:]
        if l_hash != self.hash_obj.hash(self.label):
            return ValueError("Label hashes do not match")

        return m


from samson.utilities.bytes import Bytes

class MatyasMeyerOseasConstruction(object):
    def __init__(self, initial_state, encryptor, decryptor):
        self.initial_state = initial_state
        self.encryptor = encryptor
        self.decryptor = decryptor


    def yield_state(self, message):
        block_size = len(self.initial_state)
        message = Bytes.wrap(message)

        message_blocks = message.chunk(block_size)
        last_state = self.initial_state

        for block in message_blocks:
            last_state ^= self.encryptor(last_state, block)
            yield last_state



    def hash(self, message):
        final_state = [_ for _ in self.yield_state(message)][-1]
        return final_state


from fastecdsa.point import Point
from fastecdsa.curve import P256
from samson.utilities.general import rand_bytes

class ECDHE(object):
    def __init__(self, key=None, G=P256.G):
        self.key = key or int.from_bytes(rand_bytes(), 'big')
        self.G = G


    def __repr__(self):
        return f"<ECDHE: key={self.key}, G={self.G}>"


    def __str__(self):
        return self.__repr__()


    def get_challenge(self):
        return self.key * self.G


    def derive_key(self, challenge):
        return self.key * challenge


from samson.utilities.math import gcd, lcm, mod_inv, find_prime
from samson.utilities.encoding import int_to_bytes
import random

class RSA(object):
    def __init__(self, bits, p=None, q=None, e=None):
        if p and q and not e:
            raise Exception("Argument 'e' must be set if 'p' and 'q' are set. ")

        self.e = e or 3
        phi = 0

        if p and q:
            phi = lcm(p - 1, q - 1)
            self.n = p * q

            if gcd(self.e, phi) != 1:
                raise Exception("Invalid 'p' and 'q': GCD(e, phi) != 1")
        else:
            while gcd(self.e, phi) != 1:
                #p, q = getPrime(bits // 2), getPrime(bits // 2)
                p, q = find_prime(bits // 2), find_prime(bits // 2)
                phi = lcm(p - 1, q - 1)
                self.n = p * q

        self.p = p
        self.q = q

        self.bits = bits

        self.phi = phi
        self.d = mod_inv(self.e, phi)
        self.alt_d = mod_inv(self.e, (self.p - 1) * (self.q - 1))

        self.pub = (self.e, self.n)
        self.priv = (self.d, self.n)



    def __repr__(self):
        return '<RSA: bits={}, p={}, q={}, e={}, n={}, phi={}, d={}, alt_d={}>'.format(self.bits, self.p, self.q, self.e, self.n, self.phi, self.d, self.alt_d)



    def __str__(self):
        return self.__repr__()

        
    def encrypt(self, message):
        m = int.from_bytes(message, byteorder='big')
        return pow(m, self.e, self.n)


    def decrypt(self, message):
        plaintext = pow(message, self.d, self.n)
        return int_to_bytes(plaintext, 'big')

    
    @staticmethod
    def factorize_from_shared_p(n1, n2, e):
        assert n1 != n2

        # Find shared `p`
        p = gcd(n1, n2)

        q1 = n1 // p
        q2 = n2 // p

        return (RSA(0, p=p, q=q1, e=e), RSA(0, p=p, q=q2, e=e))


    @staticmethod
    def factorize_from_faulty_crt(message, faulty_sig, e, n):
        q = gcd(pow(faulty_sig, e, n) - message, n)
        p = n // q

        return RSA(0, p=p, q=q, e=e)


    @staticmethod
    def factorize_from_d(d, e, n):
        k = d*e - 1
        p = None
        q = None

        while not p:
            g = random.randint(2, n - 1)
            t = k
            while t % 2 == 0:
                t = t // 2
                x = pow(g, t, n)
                if x > 1 and gcd(x - 1, n) > 1:
                    p = gcd(x - 1, n)
                    q = n // p
                    break
            
        return RSA(0, p=p, q=q, e=e)


from math import ceil
from samson.utilities.bytes import Bytes

class SpongeConstruction(object):
    def __init__(self, perm_func, pad_func, r, c):
        self.perm_func = perm_func
        self.pad_func = pad_func

        self.r = r
        self.c = c

        self.block_size = (self.r + 7) // 8

        self.S = [[0] * 5 for _ in range(5)]


    def __repr__(self):
        return f"<Sponge r={self.r}, c={self.c}, block_size={self.block_size}, S={self.S}, pad_func={self.pad_func}>"

    def __str__(self):
        return self.__repr__()
    

    def reset(self):
        self.S = [[0] * 5 for _ in range(5)]

    
    def absorb(self, in_bytes):
        padded = self.pad_func(in_bytes)
        
        for block in padded.chunk(self.block_size):
            curr = 0
            for y in range(5):
                for x in range(5):
                    self.S[x][y] ^= sum([byte << i * 8 for i, byte in enumerate(block[curr:curr + 8])])
                    curr += 8

            self.S = self.perm_func(self.S)

    
    def squeeze(self, amount):
        for _ in range(ceil(amount / self.block_size)):
            out = Bytes(b'')
            for y in range(5):
                for x in range(5):
                    out += Bytes(self.S[x][y]).zfill(8)[::-1]

            yield out[:self.block_size]
            self.S = self.perm_func(self.S)



from samson.utilities.math import mod_inv, lll, generate_superincreasing_seq, find_coprime
from samson.utilities.bytes import Bytes
import numpy as np
from sympy.matrices import Matrix, eye


class MerkleHellmanKnapsack(object):
    def __init__(self, priv=None, q=None, r=None, max_diff=2**20, pub_len=8):
        super_seq = generate_superincreasing_seq(pub_len + 1, max_diff, starting=Bytes.random(max(1, pub_len // 8)).int())
        self.priv = priv or super_seq[:pub_len]
        self.q = q or super_seq[-1]
        self.r = r or find_coprime(self.q, range(self.q // 4, self.q))

        self.pub = [(w * self.r) % self.q for w in self.priv]


    def __repr__(self):
        return f"<MerkleHellmanKnapsack: priv={self.priv}, pub={self.pub}, q={self.q}, r={self.r}>"
    

    def __str__(self):
        return self.__repr__()


    def encrypt(self, message):
        bin_str = ''
        for byte in message:
            bin_str += bin(byte)[2:].zfill(8)

        all_sums = []
        
        for i in range(len(bin_str) // len(self.pub)):
            byte_str = bin_str[i * len(self.pub):(i + 1) * len(self.pub)]
            all_sums.append(sum([int(byte_str[j]) * self.pub[j] for j in range(len(self.pub))]))

        return all_sums


    def decrypt(self, sums):
        r_inv = mod_inv(self.r, self.q)
        inv_sums = [(byte_sum * r_inv) % self.q for byte_sum in sums]
        plaintext = Bytes(b'')

        for inv_sum in inv_sums:
            curr = inv_sum
            bin_string = ''

            for i in range(len(self.pub) - 1, -1, -1):
                if self.priv[i] <= curr:
                    curr -= self.priv[i]
                    bin_string += '1'
                else:
                    bin_string += '0'

            plaintext += int.to_bytes(int(bin_string[::-1], 2), len(self.pub) // 8, 'big')

        return plaintext


    @staticmethod
    def recover_plaintext(ciphertext, pub):
        ident = eye(len(pub))
        pub_matrix = ident.col_join(Matrix([pub]))
        problem_matrix = pub_matrix.row_join(Matrix([[0] * len(pub) + [-ciphertext]]).T).T

        matrices = [problem_matrix.row(row) for row in range(problem_matrix.rows)]
        print(matrices)
        solution_matrix = lll(matrices, 0.99)

        for row in range(solution_matrix.rows):
            row_mat = solution_matrix.row(row)
            new_row = [item for item in row_mat if item >= 0 and item <= 1]

            if len(new_row) == len(row_mat):
                return int(''.join([str(val) for val in row_mat[:-1]]), 2)
        return solution_matrix


from samson.oracles.encryption_oracle import EncryptionOracle

class StatelessBlockEncryptionOracle(EncryptionOracle):
    def find_block_size(self):
        baseline = len(self.encrypt(b''))

        for i in range(33):
            curr_len = len(self.encrypt(('A' * i).encode()))
            if  curr_len != baseline:
                return curr_len - baseline


from samson.block_ciphers.rijndael import Rijndael
from samson.utilities.bytes import Bytes

class Poly1305(object):
    P1305 = (1 << 130) - 5
    
    def __init__(self, key, nonce, r, cipher=Rijndael):
        self.key = key
        self.nonce = nonce
        self.r = Bytes.wrap(r, byteorder='little').to_int()
        self.cipher = cipher


    def __repr__(self):
        return f"<Poly1305: key={self.key}, nonce={self.nonce}, r={self.r}, cipher={self.cipher}>"


    def __str__(self):
        return self.__repr__()

    
    def generate(self, plaintext):
        pt_chunks = [(chunk + b'\x01').zfill(17) for chunk in Bytes.wrap(plaintext, byteorder='little').chunk(16, allow_partials=True)]

        total = 0
        for chunk in pt_chunks:
            total += chunk.to_int()
            total *= self.r
        
        total %= Poly1305.P1305
        enc = self.cipher(self.key).encrypt(self.nonce)
        enc.byteorder = 'little'

        return Bytes((enc.to_int() + total) % (1 << 128), byteorder='little')


from math import ceil

def xor_buffs(buf1, buf2):
    if len(buf1) != len(buf2):
        raise Exception('Buffers must be equal length.')

    return bytearray([x ^ y for x, y in zip(buf1, buf2)])



# Stretches a `key` into a stream of `length`. Key can be shifted by `offset`.

# Examples
# >>> stretch_key(b'abc', 5)
# b'abcab'

# >>> stretch_key(b'abc', 5, offset=1)
# b'cabca'

def stretch_key(key, length, offset=0):
    offset_mod = offset % len(key)
    key_stream = (key * ceil(length / len(key)))
    complete_key = (key[-offset_mod:] + key_stream)[:length]
    return complete_key



def transpose(cipherbytes, key_size):
    return [cipherbytes[i::key_size] for i in range(key_size)]


def get_blocks(cipher, block_size=16, allow_partials=False):
    full_blocks = [cipher[i * block_size: (i + 1) * block_size] for i in range(len(cipher) // block_size)]

    left_over = len(cipher) % block_size

    if allow_partials and left_over > 0:
        all_blocks = full_blocks + [cipher[-left_over:]]
    else:
        all_blocks = full_blocks
    return all_blocks


def left_rotate(x, amount, bits=32):
    mask = 2 ** bits - 1
    x &= mask
    return ((x<<amount) | (x>>(bits-amount))) & mask


def right_rotate(x, amount, bits=32):
    mask = 2 ** bits - 1
    return ((x>>amount) | (x<<(bits-amount))) & mask


from samson.utilities.math import mod_inv
from samson.utilities.bytes import Bytes
from samson.publickey.dsa import DSA

# https://en.wikipedia.org/wiki/Elliptic_Curve_Digital_Signature_Algorithm
class ECDSA(DSA):
    def __init__(self, G, d=None):
        self.G = G
        self.q = self.G.curve.q
        self.d = d or max(1, Bytes.random(self.q.bit_length() + 7 // 8).int() % self.q)
        self.Q = self.d * self.G

    
    def __repr__(self):
        return f"<ECDSA: d={self.d}, G={self.G}, Q={self.Q}>"


    def __str__(self):
        return self.__repr__()
    

    def sign(self, H, message, k=None):
        r = 0
        s = 0

        while s == 0 or r == 0:
            k = k or max(1, Bytes.random(self.q .bit_length() + 7 // 8).int() % self.q)
            inv_k = mod_inv(k, self.q)

            z = H(message)
            z >>= max(z.bit_length() - self.q.bit_length(), 0)

            r = (k * self.G).x % self.q
            s = (inv_k * (z + self.d * r)) % self.q

        return (r, s)
    
    
    def verify(self, H, message, sig):
        (r, s) = sig
        w = mod_inv(s, self.q)

        z = H(message)
        z >>= max(z.bit_length() - self.q.bit_length(), 0)

        u_1 = (z * w) % self.q
        u_2 = (r * w) % self.q
        v = u_1 * self.G + u_2 * self.Q
        return v.x == r

    
    # # Confirmed works on ECDSA as well
    # def derive_k_from_sigs(self, H, msgA, sigA, msgB, sigB):
    #     (rA, sA) = sigA
    #     (rB, sB) = sigB
    #     assert rA == rB
        
    #     s = (sA - sB) % self.q
    #     m = (H(msgA) - H(msgB)) % self.q
    #     return mod_inv(s, self.q) * m % self.q
    
    
    # def derive_x_from_k(self, H, message, k, sig):
    #     (r, s) = sig
    #     self.d = ((s * k) - H(message)) * mod_inv(r, self.q) % self.q


import math
from samson.macs.hmac import HMAC

# https://en.wikipedia.org/wiki/HKDF
# https://tools.ietf.org/html/rfc5869
class HKDF(object):
    def __init__(self, hash_obj, desired_len):
        self.hash_obj = hash_obj
        self.desired_len = desired_len


    def __repr__(self):
        return f"<HKDF: hash_obj={self.hash_obj}, desired_len={self.desired_len}>"


    def __str__(self):
        return self.__repr__()

    
    def derive(self, key, salt, info=b''):
        prk = HMAC(key=salt, hash_obj=self.hash_obj).generate(key)
        hmac = HMAC(key=prk, hash_obj=self.hash_obj)

        new_key = b''
        t = b''
        for i in range(math.ceil(self.desired_len / (self.hash_obj.digest_size // 8))):
            t = hmac.generate(t + info + bytes([i + 1]))
            new_key += t

        return new_key[:self.desired_len]


import hashlib
from samson.utilities.math import modexp
from samson.utilities.general import rand_bytes
from samson.utilities.encoding import int_to_bytes
import codecs

NIST_prime = int.from_bytes(codecs.decode("ffffffffffffffffc90fdaa22168c234c4c6628b80dc1cd129024e088a67cc74020bbea63b139b22514a08798e3404ddef9519b3cd3a431b302b0a6df25f14374fe1356d6d51c245e485b576625e7ec6f44c42e9a637ed6b0bff5cb6f406b7edee386bfb5a899fa5ae9f24117c4b1fe649286651ece45b3dc2007cb8a163bf0598da48361c55d39a69163fa8fd24cf5f83655d23dca3ad961c62f356208552bb9ed529077096966d670c354e4abc9804f1746c08ca237327ffffffffffffffff", 'hex_codec'), 'little')


class SRPClient(object):
    def __init__(self, identity, password, g=2, k=3, N=NIST_prime):
        self.a = int.from_bytes(rand_bytes(4), 'little') % N
        self.g = g
        self.k = k
        self.A = modexp(g, self.a, N)
        self.identity = identity
        self.password = password
        self.N = N

    
    def make_request(self):
        return self.identity, self.A


    def perform_challenge(self, salt, B):
        hex_A = int_to_bytes(self.A, 'little')
        hex_B = int_to_bytes(B, 'little')

        uH = int.from_bytes(hashlib.sha256(hex_A + hex_B).digest(), 'little')
        xH = int.from_bytes(hashlib.sha256(salt + self.password).digest(), 'little')

        p1 = (B - self.k * modexp(self.g, xH, self.N))
        p2 = (self.a + uH * xH)
        cS = modexp(p1, p2, self.N)

        cK = hashlib.sha256(int_to_bytes(cS, 'little')).digest()
        return hashlib.sha256(cK + salt).digest()


    def craft_malicious_request(self):
        return self.identity, 0


    def craft_auth_bypass(self, salt):
        cK = hashlib.sha256(b'\x00' * 1).digest()
        return hashlib.sha256(cK + salt).digest()


class TokenListHandler(object):
    def __init__(self):
        self.list = []

    def reset(self):
        pass

    
    def handle_token(self, token):
        self.list.append(token)


    def finalize(self):
        return self.list


from samson.auxiliary.markov_state import MarkovState

class MarkovChainHandler(object):
    def __init__(self):
        self.chain = MarkovState(count=1, probability=1, transitions={})
        self.curr_parsed_chain = self.chain



    def reset(self):
        self.curr_parsed_chain = self.chain

    
    def handle_token(self, token):
        if token in self.curr_parsed_chain.transitions:
            self.curr_parsed_chain.transitions[token].count += 1
        else:
            self.curr_parsed_chain.transitions[token] = MarkovState(count=1, probability=0, transitions={})

        self.curr_parsed_chain = self.curr_parsed_chain.transitions[token]



    def finalize(self):
        self.chain.calculate_chain_probs()
        return self.chain


# Oracle that determines if a ciphertext has the correct padding or not. The `request` function must return a boolean indicating this.
class PaddingOracle(object):
    def __init__(self, request_func):
        self.check_padding = request_func


import struct
from samson.utilities.general import rand_bytes

def pkcs7_pad(text, block_size=16):
    padding = block_size - len(text) % block_size
    return text + struct.pack('B', padding) * padding


def pkcs7_unpad(text, block_size=16):
    last_block = text[-block_size:]
    last_byte = last_block[-1]

    original_text, padding = text[:len(text) - last_byte], last_block[-last_byte:]
    if len(padding) != last_byte or sum([last_byte != pad_char for pad_char in padding]) != 0:
        raise Exception('Invalid padding ;)')

    return original_text


def pkcs15_pad(data, key_byte_length):
    padding = rand_bytes(key_byte_length - 3 - len(data))
    return b'\x00\x02' + padding + b'\x00' + data


def md_pad(msg, fakeLen=None, byteorder='little', bit_size=64):
    length = fakeLen or len(msg)
    # append the bit '1' to the message
    padding = b'\x80'
    byte_size = bit_size // 8

    # append 0 <= k < 512 bits '0', so that the resulting message length (in bytes)
    # is congruent to 56 (mod 64)
    padding += b'\x00' * (((bit_size - byte_size) - (length + 1) % bit_size) % bit_size)

    # append length of message (before pre-processing), in bits, as 64-bit big-endian integer
    message_bit_length = length * 8
    padding += message_bit_length.to_bytes(byte_size, byteorder=byteorder)
    return msg + padding

    


from samson.utilities.bytes import Bytes

# https://en.wikipedia.org/wiki/HMAC
class HMAC(object):
    def __init__(self, key, hash_obj):
        self.key = Bytes.wrap(key)
        self.hash_obj = hash_obj

        key_prime = self.key
        if len(self.key) > self.hash_obj.block_size:
            key_prime = self.hash_obj.hash(self.key)

        self.key_prime = key_prime + b'\x00' * (self.hash_obj.block_size - len(key_prime))
        self.outer_key_pad = self.key_prime ^ Bytes(b'\x5c').stretch(self.hash_obj.block_size)
        self.inner_key_pad = self.key_prime ^ Bytes(b'\x36').stretch(self.hash_obj.block_size)


    def __repr__(self):
        return f"<HMAC: key={self.key}, key_prime={self.key_prime}, outer_key_pad={self.outer_key_pad}, inner_key_pad={self.inner_key_pad}>"

    def __str__(self):
        return self.__repr__()


    def generate(self, message):
        return self.hash_obj.hash(self.outer_key_pad + self.hash_obj.hash(self.inner_key_pad + message))


from copy import deepcopy
from samson.utilities.general import rand_bytes
from sympy.matrices import Matrix, GramSchmidt
from sympy import isprime, GF, Poly
from sympy.abc import x
import math

def gcd(a, b):
    if b == 0:
        return a
    else:
        return gcd(b, a % b)


# https://anh.cs.luc.edu/331/notes/xgcd.pdf
def xgcd(a,b):
    prevx, x = 1, 0; prevy, y = 0, 1
    while b:
        q = a//b
        x, prevx = prevx - q*x, x
        y, prevy = prevy - q*y, y
        a, b = b, a % b
    return a, b, prevx, prevy


def lcm(a, b):
    return a // gcd(a, b) * b



def mod_inv(a, n):
    """
    Calculates the modular inverse according to
    https://en.wikipedia.org/wiki/Euclidean_algorithm#Linear_Diophantine_equations
    and https://en.wikipedia.org/wiki/Extended_Euclidean_algorithm
    """

    _, b, t, _ = xgcd(a, n)

    if b > 1:
        raise Exception("'a' is not invertible")
    
    if t < 0:
        t = t + n

    return t
    

def modexp (g, u, p):
   """computes s = (g ^ u) mod p
      args are base, exponent, modulus
      (see Bruce Schneier's book, _Applied Cryptography_ p. 244)"""
   s = 1
   while u != 0:
      if u & 1:
         s = (s * g)%p
      u >>= 1
      g = (g * g)%p
   return s


# https://stackoverflow.com/questions/23621833/is-cube-root-integer
def kth_root(n,k):
    lb,ub = 0,n #lower bound, upper bound
    while lb < ub:
        guess = (lb+ub)//2
        if pow(guess,k) < n: lb = guess+1
        else: ub = guess
    return lb


def crt(residues, moduli):
    assert len(residues) == len(moduli)
    x = residues[0]
    Nx = moduli[0]

    for i in range(1, len(residues)):
        x = (mod_inv(Nx, moduli[i]) * (residues[i] - x)) * Nx + x
        Nx = Nx * moduli[i]

    return x % Nx, Nx



def legendre(a, p):
    return pow(a, (p - 1) // 2, p)


# https://crypto.stackexchange.com/questions/22919/explanation-of-each-of-the-parameters-used-in-ecc
# https://www.geeksforgeeks.org/find-square-root-modulo-p-set-2-shanks-tonelli-algorithm/
# https://rosettacode.org/wiki/Tonelli-Shanks_algorithm#Python


def tonelli(n, p):
    assert legendre(n, p) == 1, "not a square (mod p)"
    q = p - 1
    s = 0
    while q % 2 == 0:
        q //= 2
        s += 1
    if s == 1:
        return pow(n, (p + 1) // 4, p)
    for z in range(2, p):
        if p - 1 == legendre(z, p):
            break
    c = pow(z, q, p)
    r = pow(n, (q + 1) // 2, p)
    t = pow(n, q, p)
    m = s
    t2 = 0
    while (t - 1) % p != 0:
        t2 = (t * t) % p
        for i in range(1, m):
            if (t2 - 1) % p == 0:
                break
            t2 = (t2 * t2) % p
        b = pow(c, 1 << (m - i - 1), p)
        r = (r * b) % p
        c = (b * b) % p
        t = (t * c) % p
        m = i
    return r



# https://github.com/orisano/olll/blob/master/olll.py
# https://en.wikipedia.org/wiki/Lenstra%E2%80%93Lenstra%E2%80%93Lov%C3%A1sz_lattice_basis_reduction_algorithm
def lll(in_basis, delta=0.75):
    basis = deepcopy(in_basis)
    n = len(basis)
    ortho = GramSchmidt(basis)

    def mu(i, j):
        return ortho[j].dot(basis[i]) / ortho[j].dot(ortho[j])

    k = 1
    while k < n:
        for j in range(k - 1, -1, -1):
            mu_kj = mu(k, j)
            if abs(mu_kj) > 0.5:
                basis[k] = basis[k] - basis[j] * round(mu_kj)
                ortho = GramSchmidt(basis)


        if ortho[k].dot(ortho[k]) >= (delta - mu(k, k - 1)**2) * (ortho[k - 1].dot(ortho[k - 1])):
            k += 1
        else:
            basis[k], basis[k - 1] = deepcopy(basis[k - 1]), deepcopy(basis[k])
            ortho = GramSchmidt(basis)
            k = max(k - 1, 1)

    return Matrix([list(map(int, b)) for b in basis])



def generate_superincreasing_seq(length, max_diff, starting=0):
    seq = []

    last_sum = starting
    for _ in range(length):
        delta = int.from_bytes(rand_bytes(math.ceil(math.log(max_diff, 256))), 'big') % max_diff
        seq.append(last_sum + delta)
        last_sum = sum(seq)

    return seq



def find_coprime(p, search_range):
    for i in search_range:
        if gcd(p, i) == 1:
            return i



def find_prime(bits):
    rand_num = int.from_bytes(rand_bytes(math.ceil(bits / 8)), 'big')
    rand_num |= 2**(bits - 1)
    rand_num |= 1

    while not isprime(rand_num):
        rand_num += 2
    
    return rand_num


# https://en.wikipedia.org/wiki/Berlekamp%E2%80%93Massey_algorithm
def berlekamp_massey(output_list):
    n = len(output_list)
    b = [1] + [0] * (n - 1)
    c = [1] + [0] * (n - 1)

    L = 0
    m = -1

    i  = 0
    while i < n:
        out_vec = output_list[i - L:i][::-1]
        c_vec = c[1:L+i]
        d = output_list[i] + sum([s_x * c_x for s_x, c_x in zip(out_vec, c_vec)]) % 2

        if d == 1:
            t = deepcopy(c)
            p = [0] * n
            for j in range(L):
                if b[j] == 1:
                    p[j + i - m] = 1
            
            c = [(c_x + p_x) % 2 for c_x, p_x in zip(c, p)]

            if L <= i / 2:
                L = i + 1 - L
                m = i
                b = t
        
        i += 1

    return Poly(c[:L + 1], x)


def is_power_of_two(n):
    return n != 0 and (n & (n - 1) == 0)


import itertools
import struct

class Analyzer(object):
    def __init__(self, attempt_key):
        self.attempt_key = attempt_key


    def analyze(self, in_bytes):
        pass


    def build_candidates(self, in_bytes, key_length):
        all_keys = itertools.product(range(2), repeat=key_length)
        candidates = []

        for key in all_keys:
            bin_key = b''
            for i in range(max(key_length // 8, 1)):
                bin_key += struct.pack('B', int(''.join([str(curr_int) for curr_int in key[i * 8: (i + 1) * 8]]), 2))

            #, is_readable
            plaintext = self.attempt_key(bin_key, in_bytes)

            candidates.append((bin_key, plaintext))
            # if is_readable:
            #     candidates.append((bin_key, plaintext))


        return sorted(candidates, key=lambda x: self.analyze(x[1]), reverse=True)


    def select_highest_scores(self, in_list, num=1):
        return sorted(in_list, key=lambda item: self.analyze(item), reverse=True)[:num]


import codecs
from samson.utilities.math import mod_inv
from samson.utilities.general import rand_bytes
from samson.utilities.encoding import int_to_bytes
import hashlib


class DSA(object):
    def __init__(self, p=None, q=None, g=None, x=None):
        self.p = p or int.from_bytes(codecs.decode(b'800000000000000089e1855218a0e7dac38136ffafa72eda7859f2171e25e65eac698c1702578b07dc2a1076da241c76c62d374d8389ea5aeffd3226a0530cc565f3bf6b50929139ebeac04f48c3c84afb796d61e5a4f9a8fda812ab59494232c7d2b4deb50aa18ee9e132bfa85ac4374d7f9091abc3d015efc871a584471bb1', 'hex_codec'), byteorder='big')
        self.q = q or int.from_bytes(codecs.decode(b'f4f47f05794b256174bba6e9b396a7707e563c5b', 'hex_codec'), byteorder='big')
        self.g = g or int.from_bytes(codecs.decode(b'5958c9d3898b224b12672c0b98e06c60df923cb8bc999d119458fef538b8fa4046c8db53039db620c094c9fa077ef389b5322a559946a71903f990f1f7e0e025e2d7f7cf494aff1a0470f5b64c36b625a097f1651fe775323556fe00b3608c887892878480e99041be601a62166ca6894bdd41a7054ec89f756ba9fc95302291', 'hex_codec'), byteorder='big')
        
        self.x = x or int.from_bytes(rand_bytes(len(int_to_bytes(self.q))), byteorder='big') % self.q
        self.y = pow(self.g, self.x, self.p)

    
    def __repr__(self):
        return f"<DSA: p={self.p}, q={self.q}, g={self.g}, x={self.x}, y={self.y}>"


    def __str__(self):
        return self.__repr__()
    
    
    def sign(self, H, message, k=None):
        k = k or max(1, int.from_bytes(rand_bytes(len(int_to_bytes(self.q))), byteorder='big') % self.q)
        inv_k = mod_inv(k, self.q)
        r = pow(self.g, k, self.p) % self.q
        s = (inv_k * (H(message) + self.x * r)) % self.q
        return (r, s)
    
    
    def verify(self, H, message, sig):
        (r, s) = sig
        w = mod_inv(s, self.q)
        u_1 = (H(message) * w) % self.q
        u_2 = (r * w) % self.q
        v = (pow(self.g, u_1, self.p) * pow(self.y, u_2, self.p) % self.p) % self.q
        return v == r

    
    # Confirmed works on ECDSA as well
    def derive_k_from_sigs(self, H, msgA, sigA, msgB, sigB):
        (rA, sA) = sigA
        (rB, sB) = sigB
        assert rA == rB
        
        s = (sA - sB) % self.q
        m = (H(msgA) - H(msgB)) % self.q
        return mod_inv(s, self.q) * m % self.q
    
    
    # Confirmed works on ECDSA as well
    def derive_x_from_k(self, H, message, k, sig):
        (r, s) = sig
        return ((s * k) - H(message)) * mod_inv(r, self.q) % self.q


from samson.utilities.bytes import Bytes
from samson.utilities.math import mod_inv
from samson.hashes.md5 import MD5
from samson.protocols.diffie_hellman import DiffieHellman

# https://asecuritysite.com/encryption/dragonfly
class Dragonfly(object):
    def __init__(self, key, H=lambda m: MD5().hash(m), q=DiffieHellman.MODP_2048):
        self.key = key
        self.q = q
        self.A = Bytes.random(16).to_int()
        self.a = Bytes.random(16).to_int()
        self.H = H

    
    def __repr__(self):
        return f"<Dragonfly: key={self.key}, H={self.H}, A={self.A}, a={self.a}, q={self.q}>"


    def __str__(self):
        return self.__repr__()


    def get_challenge(self):
        sA = self.a + self.A

        PE = int.from_bytes(self.H(self.key)[:8], 'big')
        eA = mod_inv(pow(PE, self.A, self.q), self.q)
        return pow(PE, sA, self.q), eA


    def derive_key(self, challenge):
        PEsA, eA = challenge
        return pow(PEsA * eA, self.a, self.q)


from samson.hashes.keccak import Keccak

class SHA3(object):
    @staticmethod
    def _build_sha3(r, c, bits):
        k = Keccak(r, c, bits)

        def pad(in_bytes):
            bit_rate_bytes = (k.r + 7) // 8
            pad_len = bit_rate_bytes - (len(in_bytes) % bit_rate_bytes)

            if pad_len == 0:
                pad_len = bit_rate_bytes

            if pad_len == 1:
                return in_bytes + bytes([0x86])
            else:
                return in_bytes + bytes([0x06] + ([0] * (pad_len - 2)) + [0x80])

        
        k.pad_func = pad
        return k


    @staticmethod
    def K224():
        return SHA3._build_sha3(1152, 448, 224)


    @staticmethod
    def K256():
        return SHA3._build_sha3(1088, 512, 256)


    @staticmethod
    def K384():
        return SHA3._build_sha3(832, 768, 384)


    @staticmethod
    def K512():
        return SHA3._build_sha3(576, 1024, 512)


from samson.utilities.bytes import Bytes
from samson.utilities.math import tonelli, mod_inv
from fastecdsa.point import Point
import random

class DualEC(object):
    # P and Q are Point objects from fastecdsa
    def __init__(self, P, Q, seed):
        self.P = P
        self.Q = Q
        self.t = seed
        self.r = None


    def __repr__(self):
        return f"<DualEC: P={self.P}, Q={self.Q}, t={self.t}>"


    def __str__(self):
        return self.__repr__()


    def generate(self):
        s = (self.t * self.P).x
        self.t = s
        self.r = (s * self.Q).x

        return Bytes(int.to_bytes(self.r, 32, 'big')[2:])


    @staticmethod
    def generate_backdoor(curve):
        P = curve.G
        d = random.randint(2, curve.q)
        e = mod_inv(d, curve.q)
        Q = e * P

        return P, Q, d


    @staticmethod
    def derive_from_backdoor(P, Q, d, observed_out):
        assert len(observed_out) >= 30

        curve = P.curve

        r1 = observed_out[:30]
        r2 = observed_out[30:]

        possible_states = []

        for i in range(2**16):
            test_r1 = int.to_bytes(i, 2, 'big') + r1
            test_x = int.from_bytes(test_r1, 'big')
            try:
                y_2 = test_x ** 3 + (curve.a * test_x) + curve.b
                y = tonelli(y_2, curve.p)
                R = Point(test_x, y)
                dR = d * R
                test_r2 = dR.x * Q

                if int.to_bytes(test_r2.x, 32, 'big')[2:2 + len(r2)] == r2:
                    possible_states.append(DualEC(P, Q, dR.x))
                    print(test_r2)
            except Exception as _:
                pass

        return possible_states


from samson.prngs.mt19937 import MT19937

# We don't have to include a constant since 32-bit Mersenne Twister doesn't
# use non-idempotent constants on its right shifts.
def _untemper_right(y, bits):
    # Create a 32-bit mask with `bits` 1s at the beginning.
    # We'll shift this over the iterations to invert the temper.
    mask = (1 << bits) - 1 << 32 - bits
    shift_mod = 0

    while mask > 0:
        y ^= shift_mod

        # Get next `bits` bits of y
        # Ex: bits = 3, mask = '00011100000000000'
        # '100_010_00100100001' -> '000000_010_00000000'
        shift_mod = (y & mask) >> bits

        # Move mask right `bits` to select next bits to shift
        # Ex: bits = 3
        # 11100000000000000 -> 00011100000000000
        mask >>= bits
    return y


def _untemper_left(y, bits, constant):
    int32_mask = 0xFFFFFFFF
    mask = (1 << bits) - 1
    shift_mod = 0

    while (mask & int32_mask) > 0:
        y ^= shift_mod & constant

        # Get next `bits` bits of y
        shift_mod = (y & mask) << bits

        # Move mask right `bits` to select next bits to shift
        mask <<= bits
    return y


u, d = (11, 0xFFFFFFFF)
s, b = (7, 0x9D2C5680)
t, c = (15, 0xEFC60000)
l = 18

def _untemper(y):
    y = _untemper_right(y, l)
    y = _untemper_left(y, t, c)
    y = _untemper_left(y, s, b)
    y = _untemper_right(y, u)
    return y & d



class MT19937CloneAttack(object):
    def __init__(self):
        pass

    def execute(self, outputs):
        cloned = MT19937(0)
        cloned.state = [_untemper(output) for output in outputs][-624:]

        return cloned


from math import ceil
from samson.utilities.bytes import Bytes

class PBKDF2(object):
    def __init__(self, hash_fn, desired_len, num_iters):
        self.hash_fn = hash_fn
        self.desired_len = desired_len
        self.num_iters = num_iters


    def derive(self, password, salt):
        hash_len = len(self.hash_fn(b'', b''))
        num_blocks = ceil(self.desired_len / hash_len)

        output = Bytes(b'')
        for i in range(1, num_blocks + 1):
            xor_sum = Bytes(b'').zfill(hash_len)
            last = salt + Bytes(i).zfill(4)

            for _ in range(self.num_iters):
                last = self.hash_fn(password, last)
                xor_sum ^= last
            
            output += xor_sum
        
        return output[:self.desired_len]


from samson.utilities.bytes import Bytes
from samson.utilities.padding import pkcs7_pad

class FeistelNetwork(object):
    def __init__(self, round_func, key_schedule):
        self.round_func = round_func
        self.key_schedule = key_schedule


    def yield_encrypt(self, key, plaintext):
        half = len(plaintext) // 2
        plaintext = Bytes.wrap(plaintext)
        L_i, R_i = plaintext[:half], plaintext[half:]
        
        round_keys = list(self.key_schedule(key))

        for subkey in round_keys:
            L_i, R_i = R_i, (L_i ^ self.round_func(R_i, subkey))
            yield R_i, L_i



    def yield_decrypt(self, key, ciphertext):
        half = len(ciphertext) // 2
        ciphertext = Bytes.wrap(ciphertext)
        R_i, L_i = ciphertext[:half], ciphertext[half:]

        round_keys = list(self.key_schedule(key))[::-1]

        for subkey in round_keys:
            R_i, L_i = L_i, (R_i ^ self.round_func(L_i, subkey))
            yield L_i, R_i



    def encrypt(self, key, plaintext):
        return Bytes(b''.join(list(self.yield_encrypt(key, plaintext))[-1]))

    
    def decrypt(self, key, ciphertext):
        return Bytes(b''.join(list(self.yield_decrypt(key, ciphertext))[-1]))


from samson.utilities.manipulation import get_blocks

# https://en.wikipedia.org/wiki/Non-adjacent_form
# Takes in bytes/bytearray or int
def to_NAF(input_arg):
    if type(input_arg) is int:
        E = input_arg
    else:
        E = int.from_bytes(input_arg, 'big')
    
    z = []
    
    i = 0
    while E > 0:
        if E % 2 == 1:
            z.append(int(2 - (E % 4)))
            E -= z[-1]
        else:
            z.append(0)
        
        E /= 2
        i += 1
    return z[::-1]


def from_NAF(naf):
    total = 0
    reversed_naf = naf[::-1]
    for i in range(len(naf)):
        total += 2 ** i * reversed_naf[i]
    
    return total




def int_to_bytes(n, byteorder='big'):
    """Converts the given int n to bytes and returns them."""
    return n.to_bytes(max((n.bit_length() + 7) // 8, 1), byteorder)



def bytes_to_bitstring(input_bytes, fill=8):
    return ''.join(format(x, 'b').zfill(fill) for x in input_bytes)


# https://stackoverflow.com/questions/32675679/convert-binary-string-to-bytearray-in-python-3
def bitstring_to_bytes(bistring, byteorder='big'):
    return int(bistring, 2).to_bytes(len(bistring) // 8, byteorder=byteorder)


from samson.utilities.manipulation import get_blocks, xor_buffs

class DaviesMeyerConstruction(object):
    def __init__(self, initial_state, encryptor, decryptor):
        self.initial_state = initial_state
        self.encryptor = encryptor
        self.decryptor = decryptor


    def yield_state(self, message):
        block_size = len(self.initial_state)

        message_blocks = get_blocks(message, block_size=block_size)
        last_state = self.initial_state

        for block in message_blocks:
            last_state = xor_buffs(self.encryptor(block, last_state), last_state)
            yield last_state



    def hash(self, message):
        final_state = [_ for _ in self.yield_state(message)][-1]
        return final_state



    def generate_fixed_point(self, message):
        block_size = len(self.initial_state)
        message_blocks = get_blocks(message, block_size=block_size)

        payload = self.decryptor(message_blocks[0], b'\x00' * 16)


from samson.utilities.bytes import Bytes

class RC4(object):
    def __init__(self, key):
        self.key = key
        self.S = self.key_schedule(key)
        self.i = 0
        self.j = 0


    def __repr__(self):
        return "<RC4: key={}, S={}, i={}, j={} >".format(self.key, self.S, self.i, self.j)

    def __str__(self):
        return self.__repr__()


    def key_schedule(self, key):
        key_length = len(key)
        S = []
        for i in range(256):
            S.append(i)

        j = 0
        for i in range(256):
            j = (j + S[i] + key[i % key_length]) % 256
            S[i], S[j] = S[j], S[i]
        
        return S


    def yield_state(self, length):
        keystream = Bytes(b'')
        
        for _ in range(length):
            self.i = (self.i + 1) % 256
            self.j = (self.j + self.S[self.i]) % 256
            self.S[self.i], self.S[self.j] = self.S[self.j], self.S[self.i]
            keystream += bytes([self.S[(self.S[self.i] + self.S[self.j]) % 256]])

        return keystream


from samson.utilities.manipulation import xor_buffs
from samson.utilities.bytes import Bytes
from copy import deepcopy
import struct
import pickle
import os

import logging
log = logging.getLogger(__name__)

class XORTranspositionAttack(object):
    def __init__(self, analyzer):
        self.analyzer = analyzer


    def execute(self, ciphertexts, iterations=3):
        min_size = min([len(ciphertext) for ciphertext in ciphertexts])

        same_size_ciphers = [ciphertext[:min_size] for ciphertext in ciphertexts]
        transposed_ciphers = [bytearray(transposed) for transposed in zip(*same_size_ciphers)]
        assert [bytearray(transposed) for transposed in zip(*transposed_ciphers)] == same_size_ciphers

        log.debug("Starting initial transposition analysis")

        # Transposition analysis first (transposition)
        transposed_plaintexts = []
        for cipher in transposed_ciphers:
            all_chars = {}
            for char in range(256):
                plaintext = Bytes(struct.pack('B', char)).stretch(len(cipher)) ^ cipher

                all_chars[char] = (self.analyzer.analyze(plaintext), plaintext)

            transposed_plaintexts.append(sorted(all_chars.items(), key=lambda kv: kv[1][0], reverse=True)[0][1][1])


        retransposed_plaintexts = [bytearray(transposed) for transposed in zip(*transposed_plaintexts)]

        log.debug("Starting full-text analysis on retransposed text")

        # Clean up with a character-by-character, higher-context analysis (retransposed)
        for j in range(iterations):
            log.debug("Starting iteration {}/{}".format(j, iterations))
            differential_mask = bytearray()

            for i in range(min_size):
                all_chars = {}

                for char in range(256):
                    full_text_analyses = []

                    frames = []
                    for curr_cipher in retransposed_plaintexts:
                        cipher_copy = deepcopy(curr_cipher)
                        cipher_copy[i] = ord(Bytes(struct.pack('B', char)) ^ struct.pack('B', curr_cipher[i]))

                        preprocessed_frame = self.analyzer.preprocess(cipher_copy)
                        frames.append(preprocessed_frame)
                        full_text_analyses.append(self.analyzer.analyze(cipher_copy))

                    all_chars[char] = (sum(full_text_analyses), char)

                best_char = sorted(all_chars.items(), key=lambda kv: kv[1][0], reverse=True)[0][1][1]
                differential_mask += struct.pack('B', best_char)
                
            retransposed_plaintexts = [xor_buffs(cipher, differential_mask) for cipher in retransposed_plaintexts]

        return retransposed_plaintexts


from samson.utilities.math import find_prime, lcm, mod_inv, gcd
from samson.utilities.bytes import Bytes


class Paillier(object):
    def __init__(self, p=None, q=None):
        self.p = p or find_prime(512)
        self.q = q or find_prime(512)
        self.n = self.p * self.q
        #self.lamb = lcm(p - 1, q - 1)

        #n_sqr = self.n ** 2
        #max_num = (n_sqr).bit_length() + 7 // 8
        #self.g = Bytes.random(max_num).to_int()
        self.phi = (self.p - 1) * (self.q - 1)
        self.g = self.n + 1
        self.priv = mod_inv(self.phi, self.n)

        #self.priv = mod_inv(self.L(pow(self.g, self.lamb, n_sqr)), self.n)


    def L(self, x):
        return (x - 1) // self.n


    def __repr__(self):
        return f"<Paillier: priv={self.priv}, p={self.p}, q={self.q}, n={self.n}, g={self.g}, phi={self.phi}>"


    def __str__(self):
        return self.__repr__()



    def encrypt(self, plaintext):
        m = Bytes.wrap(plaintext).to_int()
        assert m < self.n

        r = Bytes.random(self.n.bit_length() + 7 // 8).to_int()
        while gcd(r, self.n) != 1:
            r = Bytes.random(self.n.bit_length() + 7 // 8).to_int()

        n_sqr = self.n ** 2
        return pow(self.g, m, n_sqr) * pow(r, self.n, n_sqr)

    
    def decrypt(self, ciphertext):
        return Bytes((self.L(pow(ciphertext, self.phi, self.n ** 2)) * self.priv) % self.n)



from samson.utilities.bytes import Bytes



class CBCIVKeyEquivalenceAttack(object):
    def __init__(self, oracle, block_size=16):
        self.oracle = oracle
        self.block_size = block_size


    def execute(self, ciphertext):
        edited_cipher = Bytes.wrap(bytes(ciphertext))

        edited_cipher[self.block_size:self.block_size * 2] = b'\x00' * self.block_size
        edited_cipher[self.block_size * 2:self.block_size * 3] = edited_cipher[:self.block_size]

        recovered_bytes = self.oracle.request(bytes(edited_cipher))
        blocks = Bytes(recovered_bytes).chunk(self.block_size)

        derived_iv = blocks[0] ^ blocks[2]
        return derived_iv